''' Funciones para cargar o crear una especie en la base de datos '''

import pandas as pd
from typing import Optional, Union, Dict

# SQLAlchemy
import sqlalchemy as sa
import sqlalchemy.orm as orm

# Modelos
from modelos import Conexion, Especie
from modelos import GridGeo64km, GridGeo32km, GridGeo16km, GridGeo8km

# Variables globales de conexión a la base de datos
MOTOR_SQL: sa.engine.Engine
SESION: orm.Session

def usar_conexion(conexion: Conexion):
	''' Usa el motor y la sesión de una conexión a la base de datos '''

	global MOTOR_SQL, SESION
	MOTOR_SQL = conexion.motor_sql
	SESION = conexion.sesion

def cargar_especie(grupo_edad: str, sexo: str, tipo: str) -> Optional[Especie]:
	''' Carga una especie desde la base de datos '''

	conjunto, clasificacion = tipo.split(' ')

	especie: Optional[Especie] = SESION \
		.query(Especie) \
		.filter(Especie.clasificacion == clasificacion) \
		.filter(Especie.grupo_edad == grupo_edad) \
		.filter(Especie.conjunto == conjunto) \
		.filter(Especie.sexo == sexo) \
		.first()

	return especie

def crear_especie(grupo_edad: str, sexo: str, tipo: str):
	''' Crea una especie y la inserta en la base de datos '''

	conjunto, clasificacion = tipo.split(' ')

	especie = Especie \
	(
		grupo_edad = grupo_edad,
		sexo = sexo,
		conjunto = conjunto,
		clasificacion = clasificacion,
		tipo = tipo,
		validacion = 2,
		cells_64km = [],
		cells_32km = [],
		cells_16km = [],
		cells_8km = []
	)

	SESION.add(especie)
	SESION.flush()

	return especie

def cargar_o_crear_especie(grupo_edad, sexo, tipo: str):
	''' Carga una especie de la base de datos, o la crea si no existe '''

	especie = cargar_especie(grupo_edad, sexo, tipo)
	return crear_especie(grupo_edad, sexo, tipo) if especie is None else especie

def agregar_celdas_especie(registros: pd.DataFrame, especie: pd.Series, resolucion):
	'''
		Dados los registros de una especie, obtiene las celdas de cierta resolución
		y las agrega al conjunto de celdas de esa especie conservando unicamente
		las celdas únicas
	'''

	nueva_especie: pd.DataFrame = registros[registros['spid'] == especie['spid']]
	nuevas_celdas = nueva_especie[f'gridid_{resolucion}km'].unique().tolist()
	celdas = especie[f'cells_{resolucion}km'] + nuevas_celdas
	return pd.unique(celdas).tolist()

def actualizar_celdas_especie(registros: pd.DataFrame, especie: pd.Series):
	''' Actualiza las celdas de todas las resoluciones de una especie con los nuevos registros '''

	SESION.query(Especie) \
		.filter(Especie.spid == especie['spid']) \
		.update \
		({
			Especie.cells_64km: agregar_celdas_especie(registros, especie, 64),
			Especie.cells_32km: agregar_celdas_especie(registros, especie, 32),
			Especie.cells_16km: agregar_celdas_especie(registros, especie, 16),
			Especie.cells_8km : agregar_celdas_especie(registros, especie, 8),
		})

	SESION.flush()
	return especie

def actualizar_celdas_regiones(especie: Especie):
	''' Actualiza las celdas de todas las regiones de una especie '''

	resoluciones: Dict[int, Union[GridGeo64km, GridGeo32km, GridGeo16km, GridGeo8km]] = \
	{
		64: GridGeo64km,
		32: GridGeo32km,
		16: GridGeo16km,
		8 : GridGeo8km
	}

	regiones = SESION.query(GridGeo64km.footprint_region).all()

	for region, in regiones:
		for resolucion, GridGeo in resoluciones.items():
			nuevas_celdas: sa.Column = getattr(Especie, f'cells_{resolucion}km')
			celdas_region: sa.Column = getattr(Especie, f'cells_{resolucion}km_{region}')
			interseccion: sa.Column = nuevas_celdas.op('&')(GridGeo.cells).label('interseccion')

			# Se obtiene la intersección de las celdas de la especie con las celdas de la región
			subconsulta: orm.Query = SESION.query(Especie.spid, interseccion) \
				.join(GridGeo, GridGeo.footprint_region == region) \
				.filter(Especie.spid == especie.spid) \
				.subquery()

			# Se actualiza la especie con la intersección de las celdas
			SESION.query(Especie) \
				.filter(Especie.spid == subconsulta.c.spid) \
				.update({celdas_region: subconsulta.c.interseccion}, synchronize_session = False)

	SESION.flush()
	return especie