''' Funciones para cargar o crear una geolocalización en la base de datos '''

from typing import Optional

# SQLAlchemy
import sqlalchemy as sa
import sqlalchemy.orm as orm
from geoalchemy2.functions import ST_GeomFromText

# Modelos
from modelos import Conexion
from modelos import Aoi, Grid8km, Geolocalizacion
from modelos import GridStateAoi, GridMunAoi, GridAgebAoi

# Variables globales de conexión a la base de datos
MOTOR_SQL: sa.engine.Engine
SESION: orm.Session

def usar_conexion(conexion: Conexion):
	''' Usa el motor y la sesión de una conexión a la base de datos '''

	global MOTOR_SQL, SESION
	MOTOR_SQL = conexion.motor_sql
	SESION = conexion.sesion

def cargar_geolocalizacion(longitud: float, latitud: float) -> Optional[Geolocalizacion]:
	''' Carga una geolocalización desde la base de datos '''

	the_geom = f'SRID=4326;POINT({longitud} {latitud})'

	geolocalizacion = SESION \
		.query(Geolocalizacion) \
		.filter(Geolocalizacion.the_geom.ST_Equals(the_geom)) \
		.first()

	return geolocalizacion

def crear_geolocalizacion(longitud: float, latitud: float):
	''' Crea una geolocalización y la inserta en la base de datos '''

	the_geom = f'SRID=4326;POINT({longitud} {latitud})'
	geom_m = f'SRID=900913;POINT({longitud} {latitud})'

	# Se carga el campo 'gid' de la tabla 'aoi'
	datos_aoi: Optional[Aoi] = SESION \
		.query(Aoi.gid) \
		.filter(Aoi.geom.ST_Intersects(ST_GeomFromText(the_geom))) \
		.first()

	# Se cargan las llaves foráneas 'gridid_8km', 'gridid_16km', 'gridid_32km' y 'gridid_64km' de la tabla 'grid_8km_aoi'
	datos_grid: Optional[Grid8km] = SESION \
		.query(Grid8km.gridid_8km, Grid8km.gridid_16km, Grid8km.gridid_32km, Grid8km.gridid_64km) \
		.filter(Grid8km.the_geom.ST_Intersects(ST_GeomFromText(the_geom))) \
		.first()

	# Se carga la llave foránea 'gridid_statekm' de la tabla 'grid_statekm_aoi'
	datos_statekm: Optional[GridStateAoi] = SESION \
		.query(GridStateAoi.gridid_statekm) \
		.filter(GridStateAoi.the_geom.ST_Intersects(ST_GeomFromText(the_geom))) \
		.first()

	# Se carga la llave foránea 'gridid_munkm' de la tabla 'grid_munkm_aoi'
	datos_munkm: Optional[GridMunAoi] = SESION \
		.query(GridMunAoi.gridid_munkm) \
		.filter(GridMunAoi.the_geom.ST_Intersects(ST_GeomFromText(the_geom))) \
		.first()

	# Se carga la llave foránea 'gridid_agebkm' de la tabla 'grid_agebkm_aoi'
	datos_agebkm: Optional[GridAgebAoi] = SESION \
		.query(GridAgebAoi.gridid_agebkm) \
		.filter(GridAgebAoi.the_geom.ST_Intersects(ST_GeomFromText(the_geom))) \
		.first()

	# Función utilizada para asegurar que las llaves foráneas sean siempre valores enteros
	def llave_foranea(resultado: Optional[sa.engine.Row], campo: str):
		try:
			return int(resultado[campo])
		except:
			return 0

	# Se crea el registro en la tabla 'geo_snib'
	geolocalizacion = Geolocalizacion \
	(
		longitud = longitud,
		latitud =  latitud,
		the_geom = the_geom,
		geom_m =   geom_m,

		# Campos obtenidos de la tabla 'aoi'
		gid = llave_foranea(datos_aoi, 'gid'),

		# Campos obtenidos de la tabla 'grid_8km_aoi'
		gridid_8km =  llave_foranea(datos_grid, 'gridid_8km'),
		gridid_16km = llave_foranea(datos_grid, 'gridid_16km'),
		gridid_32km = llave_foranea(datos_grid, 'gridid_32km'),
		gridid_64km = llave_foranea(datos_grid, 'gridid_64km'),

		# Campos obtenidos de la tablas 'grid_statekm_aoi', 'grid_munkm_aoi' y 'grid_agebkm_aoi'
		gridid_statekm = llave_foranea(datos_statekm, 'gridid_statekm'),
		gridid_munkm =   llave_foranea(datos_munkm,   'gridid_munkm'),
		gridid_agebkm =  llave_foranea(datos_agebkm,  'gridid_agebkm')
	)

	# Se guarda el registro en la base de datos
	SESION.add(geolocalizacion)
	SESION.flush()

	return geolocalizacion

def cargar_o_crear_geolocalizacion(longitud: float, latitud: float):
	''' Carga una geolocalización desde la base de datos o la crea si no existe '''

	geolocation = cargar_geolocalizacion(longitud, latitud)
	return crear_geolocalizacion(longitud, latitud) if geolocation is None else geolocation