''' Contiene las clases de conexión y los modelos asociados a las base de datos '''

from typing import Optional
from pandas import Timestamp

# SQLAlchemy
import sqlalchemy as sa
import geoalchemy2 as ga2
import sqlalchemy.orm as orm
from sqlalchemy.pool import NullPool

# Clase base de SQLAlchemy
PostgreSQL = orm.declarative_base()

# ------------------------ Conexión a la base de datos ------------------------

class Conexion:
	''' Representa una conexión genérica a una base de datos '''

	def __init__(self, cadena_conexion: str, metadata: sa.MetaData):
		# Se conecta a la base de datos y se crea la sesión de SQLAlchemy
		self.motor_sql = sa.create_engine(cadena_conexion, echo = False, poolclass = NullPool)
		self.sesion = orm.Session(bind = self.motor_sql, autoflush = False)

		# Inicializa la base de datos correspondiente con los modelos de SQLAlchemy
		metadata.create_all(self.motor_sql)

	@property
	def enlace(self):
		''' Regresa el enlace de la conexión de la transacción actual '''

		return self.sesion.connection()

	def cerrar(self):
		''' Cierra la sesión de SQLAlchemy '''

		self.sesion.close()

class ConexionPostgreSQL(Conexion):
	''' Representa una conexión a una base de datos de PostgreSQL '''

	def __init__ \
	(
		self,
		host: str,
		puerto: int,
		usuario: str,
		clave: str,
		base_de_datos: str
	):
		# Se crea la cadena de conexión a la base de datos de PostgreSQL y se hace la conexión
		cadena_conexion = f'postgresql://{usuario}:{clave}@{host}:{puerto}/{base_de_datos}'
		super().__init__(cadena_conexion, PostgreSQL.metadata)

# ---------------------------- Modelos de PostgreSQL --------------------------

class Aoi(PostgreSQL):
	__tablename__ = 'aoi'
	gid =     sa.Column(sa.Integer, primary_key = True)
	country = sa.Column(sa.String(32))
	geom =    sa.Column(ga2.Geometry)
	fgid =    sa.Column(sa.Integer)

class SnibStateView(PostgreSQL):
	__tablename__ = 'snib_grid_statekm'
	id =             sa.Column(sa.Integer, primary_key = True) # Columna no existente en la tabla original
	spid =           sa.Column(sa.Integer)
	gridid_statekm = sa.Column(sa.Integer)

class SnibMunView(PostgreSQL):
	__tablename__ = 'snib_grid_munkm'
	id =           sa.Column(sa.Integer, primary_key = True) # Columna no existente en la tabla original
	spid =         sa.Column(sa.Integer)
	gridid_munkm = sa.Column(sa.Integer)

class SnibAgebView(PostgreSQL):
	__tablename__ = 'snib_grid_agebkm'
	id =            sa.Column(sa.Integer, primary_key = True) # Columna no existente en la tabla original
	spid =          sa.Column(sa.Integer)
	gridid_agebkm = sa.Column(sa.Integer)

class GridStateAoi(PostgreSQL):
	__tablename__ = 'grid_statekm_aoi'
	gridid_statekm = sa.Column(sa.Integer, primary_key = True)
	the_geom =       sa.Column(ga2.Geometry('Polygon', 4326))

class GridMunAoi(PostgreSQL):
	__tablename__ = 'grid_munkm_aoi'
	gridid_munkm = sa.Column(sa.Integer, primary_key = True)
	the_geom =     sa.Column(ga2.Geometry('Polygon', 4326))

class GridAgebAoi(PostgreSQL):
	__tablename__ = 'grid_agebkm_aoi'
	gridid_agebkm = sa.Column(sa.Integer, primary_key = True)
	the_geom =      sa.Column(ga2.Geometry('Polygon', 4326))

class GridGeo64km(PostgreSQL):
	__tablename__ = 'grid_geojson_64km_aoi'
	footprint_region = sa.Column(sa.Integer, primary_key = True)
	cells =            sa.Column(sa.ARRAY(sa.Integer))

class GridGeo32km(PostgreSQL):
	__tablename__ = 'grid_geojson_32km_aoi'
	footprint_region = sa.Column(sa.Integer, primary_key = True)
	cells =            sa.Column(sa.ARRAY(sa.Integer))

class GridGeo16km(PostgreSQL):
	__tablename__ = 'grid_geojson_16km_aoi'
	footprint_region = sa.Column(sa.Integer, primary_key = True)
	cells =            sa.Column(sa.ARRAY(sa.Integer))

class GridGeo8km(PostgreSQL):
	__tablename__ = 'grid_geojson_8km_aoi'
	footprint_region = sa.Column(sa.Integer, primary_key = True)
	cells =            sa.Column(sa.ARRAY(sa.Integer))

class Grid64km(PostgreSQL):
	__tablename__ = 'grid_64km_aoi'
	gridid_64km = sa.Column(sa.Integer, primary_key = True)
	gcol =        sa.Column(sa.Integer)
	grow =        sa.Column(sa.Integer)
	geom =        sa.Column(ga2.Geometry)
	the_geom =    sa.Column(ga2.Geometry('Polygon', 4326))
	small_geom =  sa.Column(ga2.Geometry('Polygon', 4326))

class Grid32km(PostgreSQL):
	__tablename__ = 'grid_32km_aoi'
	gridid_32km = sa.Column(sa.Integer, primary_key = True)
	gcol =        sa.Column(sa.Integer)
	grow =        sa.Column(sa.Integer)
	geom =        sa.Column(ga2.Geometry)
	the_geom =    sa.Column(ga2.Geometry('Polygon', 4326))
	small_geom =  sa.Column(ga2.Geometry('Polygon', 4326))

	# Llaves foráneas
	gridid_64km = sa.Column(sa.Integer)

class Grid16km(PostgreSQL):
	__tablename__ = 'grid_16km_aoi'
	gridid_16km = sa.Column(sa.Integer, primary_key = True)
	gcol =        sa.Column(sa.Integer)
	grow =        sa.Column(sa.Integer)
	geom =        sa.Column(ga2.Geometry)
	the_geom =    sa.Column(ga2.Geometry('Polygon', 4326))
	small_geom =  sa.Column(ga2.Geometry('Polygon', 4326))

	# Llaves foráneas
	gridid_64km = sa.Column(sa.Integer)
	gridid_32km = sa.Column(sa.Integer)

class Grid8km(PostgreSQL):
	__tablename__ = 'grid_8km_aoi'
	gridid_8km = sa.Column(sa.Integer, primary_key = True)
	gcol =       sa.Column(sa.Integer)
	grow =       sa.Column(sa.Integer)
	geom =       sa.Column(ga2.Geometry)
	the_geom =   sa.Column(ga2.Geometry('Polygon', 4326))
	small_geom = sa.Column(ga2.Geometry('Polygon', 4326))

	# Llaves foráneas
	gridid_64km = sa.Column(sa.Integer)
	gridid_32km = sa.Column(sa.Integer)
	gridid_16km = sa.Column(sa.Integer)

class Geolocalizacion(PostgreSQL):
	__tablename__ = 'geo_snib'
	geoid =    sa.Column(sa.Integer, primary_key = True)
	longitud = sa.Column(sa.Float)
	latitud =  sa.Column(sa.Float)
	the_geom = sa.Column(ga2.Geometry('POINT', 4326))
	geom_m =   sa.Column(ga2.Geometry('POINT', 900913))

	# Llaves foráneas
	gid =            sa.Column(sa.Integer)
	gridid_64km =    sa.Column(sa.Integer)
	gridid_32km =    sa.Column(sa.Integer)
	gridid_16km =    sa.Column(sa.Integer)
	gridid_8km  =    sa.Column(sa.Integer)
	gridid_statekm = sa.Column(sa.Integer)
	gridid_munkm =   sa.Column(sa.Integer)
	gridid_agebkm =  sa.Column(sa.Integer)

class Especie(PostgreSQL):
	__tablename__ = 'sp_snib'
	spid =          sa.Column(sa.Integer, primary_key = True)
	grupo_edad =    sa.Column('reinovalido', sa.String(100))
	sexo =          sa.Column('phylumdivisionvalido', sa.String(100))
	conjunto =      sa.Column('generovalido', sa.String(100))
	clasificacion = sa.Column('especieepiteto', sa.String(100))
	tipo =          sa.Column('especievalidabusqueda', sa.String(200))
	validacion =    sa.Column('validadoterceros', sa.SmallInteger)

	# Campos temporales usados para calcular el resto de regiones
	cells_64km = sa.Column(sa.ARRAY(sa.Integer))
	cells_32km = sa.Column(sa.ARRAY(sa.Integer))
	cells_16km = sa.Column(sa.ARRAY(sa.Integer))
	cells_8km =  sa.Column(sa.ARRAY(sa.Integer))

	# Campos asociados a la region '1'
	cells_64km_1 =    sa.Column(sa.ARRAY(sa.Integer))
	cells_32km_1 =    sa.Column(sa.ARRAY(sa.Integer))
	cells_16km_1 =    sa.Column(sa.ARRAY(sa.Integer))
	cells_8km_1 =     sa.Column(sa.ARRAY(sa.Integer))
	cells_statekm_1 = sa.Column(sa.ARRAY(sa.String))
	cells_munkm_1 =   sa.Column(sa.ARRAY(sa.String))
	cells_agebkm_1 =  sa.Column(sa.ARRAY(sa.String))

class Registro(PostgreSQL):
	__tablename__ = 'snib'
	id =           sa.Column('idejemplar', sa.String(64), primary_key = True)
	grupo_edad =   sa.Column('reinovalido', sa.String(100))
	sexo =         sa.Column('phylumdivisionvalido', sa.String(100))
	conjunto =     sa.Column('generovalido', sa.String(100))
	tipo =         sa.Column('especievalida', sa.String(200))
	copia_tipo =   sa.Column('especievalidabusqueda', sa.String(200))
	nacionalidad = sa.Column('proyecto', sa.String(100))

	# Datos geográficos
	longitud =   sa.Column(sa.Float)
	latitud =    sa.Column(sa.Float)
	pais_geo =   sa.Column('paismapa', sa.String(100))
	estado_geo = sa.Column('estadomapa', sa.String(100))
	the_geom =   sa.Column(ga2.Geometry('POINT', 4326))

	# Fecha de ingreso / síntomas / defunción
	dia_ingreso =       sa.Column('diadeterminacion', sa.Integer)
	mes_ingreso =       sa.Column('mesdeterminacion', sa.Integer)
	anio_ingreso =      sa.Column('aniodeterminacion', sa.SmallInteger)
	dia_sintomas_def =  sa.Column('diacolecta', sa.Integer)
	mes_sintomas_def =  sa.Column('mescolecta', sa.Integer)
	anio_sintomas_def = sa.Column('aniocolecta', sa.SmallInteger)

	# Llaves foráneas
	gid =            sa.Column(sa.Integer)
	spid =           sa.Column(sa.Integer)
	geoid =          sa.Column(sa.Integer)
	gridid_64km =    sa.Column(sa.Integer)
	gridid_32km =    sa.Column(sa.Integer)
	gridid_16km =    sa.Column(sa.Integer)
	gridid_8km =     sa.Column(sa.Integer)
	gridid_statekm = sa.Column(sa.Integer)
	gridid_munkm =   sa.Column(sa.Integer)
	gridid_agebkm =  sa.Column(sa.Integer)

	# Llaves foráneas
	gid =         sa.Column(sa.Integer)
	gridid_64km = sa.Column(sa.Integer)
	gridid_32km = sa.Column(sa.Integer)
	gridid_16km = sa.Column(sa.Integer)
	gridid_8km  = sa.Column(sa.Integer)

# ------------------------------ Modelos de Kafka -----------------------------

class RegistroKafka:
	''' Modelo que representa un registro del tópico de Kafka de casos de COVID-19 '''

	FECHA_ACTUALIZACION: Optional[str] = None
	ID_REGISTRO: Optional[str] = None
	ORIGEN: Optional[int] = None
	SECTOR: Optional[int] = None
	ENTIDAD_UM: Optional[int] = None
	SEXO: Optional[int] = None
	ENTIDAD_NAC: Optional[int] = None
	ENTIDAD_RES: Optional[int] = None
	MUNICIPIO_RES: Optional[int] = None
	TIPO_PACIENTE: Optional[int] = None
	FECHA_INGRESO: Optional[str] = None
	FECHA_SINTOMAS: Optional[str] = None
	FECHA_DEF: Optional[str] = None
	INTUBADO: Optional[int] = None
	NEUMONIA: Optional[int] = None
	EDAD: Optional[int] = None
	NACIONALIDAD: Optional[int] = None
	EMBARAZO: Optional[int] = None
	HABLA_LENGUA_INDIG: Optional[int] = None
	INDIGENA: Optional[int] = None
	DIABETES: Optional[int] = None
	EPOC: Optional[int] = None
	ASMA: Optional[int] = None
	INMUSUPR: Optional[int] = None
	HIPERTENSION: Optional[int] = None
	OTRA_COM: Optional[int] = None
	CARDIOVASCULAR: Optional[int] = None
	OBESIDAD: Optional[int] = None
	RENAL_CRONICA: Optional[int] = None
	TABAQUISMO: Optional[int] = None
	OTRO_CASO: Optional[int] = None
	TOMA_MUESTRA_LAB: Optional[int] = None
	RESULTADO_LAB: Optional[int] = None
	TOMA_MUESTRA_ANTIGENO: Optional[int] = None
	RESULTADO_ANTIGENO: Optional[int] = None
	CLASIFICACION_FINAL: Optional[int] = None
	MIGRANTE: Optional[int] = None
	PAIS_NACIONALIDAD: Optional[str] = None
	PAIS_ORIGEN: Optional[str] = None
	UCI: Optional[int] = None
	TYPE: Optional[str] = None

	def __init__(self, **kwargs):
		for campo, valor in kwargs.items():
			setattr(self, campo, valor)

	@property
	def fecha_actualizacion(self):
		return Timestamp(self.FECHA_ACTUALIZACION)

	@staticmethod
	def desde_diccionario(diccionario: dict, _):
		''' Método requerido por Kafka para poder deserializar un objeto de este tipo '''

		if diccionario is None:
			return None
		return RegistroKafka(**diccionario)